# $Id: aws.rb,v 1.6 2008/03/25 00:50:07 ianmacd Exp $
#

module Amazon
  module AWS

    require 'uri'
    require 'amazon'
    require 'amazon/aws/cache'

    NAME = '%s/%s' % [ Amazon::NAME, 'AWS' ]
    VERSION = '0.0.1'
    USER_AGENT = '%s %s' % [ NAME, VERSION ]

    # Default Associate IDs to use per locale.
    #
    DEF_ASSOC  = {
      'ca' => 'caliban-20',
      'de' => 'calibanorg0a-21',
      'fr' => 'caliban08-21',
      'jp' => 'calibanorg-20',
      'uk' => 'caliban-21',
      'us' => 'calibanorg-20'
    }

    # Service name and version for AWS.
    #
    SERVICE = { 'Service' => 'AWSECommerceService',
		'Version' => '2008-03-03'
    }

    # Maximum number of 301 and 302 HTTP responses to follow, should Amazon
    # later decide to change the location of the service.
    #
    MAX_REDIRECTS = 3

    # Maximum number of results pages that can be retrieved for a given
    # search.
    #
    MAX_PAGES = 400


    class Endpoint

      attr_reader :host, :path

      def initialize(endpoint)
	uri = URI.parse( endpoint )
	@host = uri.host
	@path = uri.path
      end
    end

    ENDPOINT = {
      'ca' => Endpoint.new( 'http://ecs.amazonaws.ca/onca/xml' ),
      'de' => Endpoint.new( 'http://ecs.amazonaws.de/onca/xml' ),
      'fr' => Endpoint.new( 'http://ecs.amazonaws.fr/onca/xml' ),
      'jp' => Endpoint.new( 'http://ecs.amazonaws.jp/onca/xml' ),
      'uk' => Endpoint.new( 'http://ecs.amazonaws.co.uk/onca/xml' ),
      'us' => Endpoint.new( 'http://ecs.amazonaws.com/onca/xml' )
    }

    # Get page, but throw exception if there's an HTTP error.
    #
    def AWS.get_page(request, query)

      url = ENDPOINT[request.locale].path + query
      cache_url = ENDPOINT[request.locale].host + url

      # Check for cached page and return that if it's there.
      #
      if request.cache && request.cache.cached?( cache_url )
        body = request.cache.get_cached( cache_url )
        return body if body
      end

      # Get the existing connection. If there isn't one, force a new one.
      #
      conn = request.conn || request.reconnect.conn
      user_agent = request.user_agent

      Amazon::dprintf( 'Fetching http://%s%s ...', conn.address, url )

      begin
        response = conn.get( url, { 'user-agent', user_agent } )

      # If we've pulled and processed a lot of pages from the cache (or
      # just not passed by here recently), the HTTP connection to the server
      # will probably have timed out.
      #
      rescue Errno::ECONNRESET
	conn = request.reconnect.conn
	retry
      end

      redirects = 0
      while response.key? 'location'
        if ( redirects += 1 ) > MAX_REDIRECTS
          raise HTTPError, "More than #{MAX_REDIRECTS} redirections"
        end

        old_url = url
        url = URI.parse( response['location'] )
        url.scheme = old_url.scheme unless url.scheme
        url.host = old_url.host unless url.host
        Amazon::dprintf( 'Following HTTP %s to %s ...', response.code, url )
        response = Net::HTTP::start( url.host ).
      	       get( url.path,{ 'user-agent', user_agent } )
      end

      if response.code != '200'
        raise HTTPError, "HTTP response code #{response.code}"
      end

      # Cache the page if we're using a cache.
      #
      if request.cache
        request.cache.cache( cache_url, response.body )
      end

      response.body
    end


    def AWS.assemble_query(items)
      query = ''

      # We must sort the items into an array to get reproducible ordering
      # of the query parameters. Otherwise, URL caching would not work.
      #
      items.sort.each do |k, v|
        query << '&%s=%s' % [ k, Amazon.url_encode( v.to_s ) ]
      end

      # Replace initial ampersand with question-mark.
      #
      query[0] = '?'

      query
    end


    class AWSObject < AmazonObject

      def instantiate_attributes(xml)
	root = xml

	# Make and populate a hash with the frequency of each attribute.
	#
	freq = Hash.new( 0 )
	root.each { |e| freq[e.name] += 1 }

	# Now create instance variables for each attribute. Multi-value
	# attributes get an array, single-value attributes a scalar, and
	# ListPrice and PackageDimensions new objects from those classes.
	#
	freq.each do |at, f|
	  iv = Amazon.uncamelise( at )

	  if f > 1
	    value = root.get_elements( at ).collect { |e| e.text }

	  elsif root.elements[at].has_elements?

	    value = case at
	    when 'Price', 'ListPrice'
	      Price.new( root.elements[at] )
	    when 'ItemDimensions'
	      ItemDimensions.new( root.elements[at] )
	    when 'PackageDimensions'
	      PackageDimensions.new( root.elements[at] )
	    when 'Languages'
	      LanguageSet.new( root.elements[at] )
	    when 'Seller'
	      Seller.new( root.elements[at] )
	    else
	      # FIXME: Shouldn't reach here.
	      raise "unknown attribute (#{at}) with child elements"
	    end

	  else
	    value = root.elements[at].text
	  end

	  instance_variable_set( "@#{iv}".to_sym, value )
	end

      end

    end


    # An exception generator class.
    #
    class AWSError

      attr_reader :exception, :msg

      def initialize(xml)
	error_code = xml.elements['Code'].text
	error_code = error_code.sub( /^AWS.*\./, '' )
        @msg = xml.elements['Message'].text

	# Ugh. Maybe there's a better way of dynamically creating an
	# exception class.
	#
	exception_str = 'class %s < StandardError; end' % [ error_code ]
	Amazon::AWS.module_eval exception_str
	@exception = Object.const_get( error_code )
      end
    end


    # This is the base class of all AWS operations.
    #
    class Operation

      OPERATIONS = %w[
	BrowseNodeLookup ItemLookup ItemSearch ListSearch SellerListingSearch
      ]

      # These are the valid search parameters that can be used with
      # ItemSearch.
      #
      PARAMETERS = %w[
	Actor		Artist	      AudienceRating	Author
	Brand		BrowseNode    City Composer	Conductor
	Director	Keywords      Manufacturer	MusicLabel
	Neighborhood	Orchestra     Power		Publisher
	TextStream	Title
      ]

      OPT_PARAMETERS = %w[
	Availability Condition MinimumPrice MaximumPrice
	MerchantId OfferStatus Sort
      ]

      ALL_PARAMETERS = PARAMETERS + OPT_PARAMETERS

      attr_reader :kind, :params

      def initialize(parameters)

	op_kind = self.class.to_s.sub( /^.*::/, '' )
	raise "Bad operation: #{op_kind}" unless OPERATIONS.include? op_kind
	raise 'Too many parameters' if parameters.size > 10

	@kind = op_kind
	@params = { 'Operation' => op_kind }.merge( parameters )
      end


      def batch_parameters( params, *b_params )

	unless b_params.empty?
	  class_str = self.class.to_s.sub( /^.+::/, '' )
	  all_parameters = [ params ].concat( b_params )
	  params = {}

	  all_parameters.each_with_index do |hash, index|
	    hash.each do |tag, val|
	      shared_param = '%s.%d.%s' % [ class_str, index + 1, tag ]
	      params[shared_param] = val
	    end
	  end

	end

	params
      end

      def parameter_check(parameters)
	parameters.each_key do |key|
	  raise "bad parameter: #{key}" unless ALL_PARAMETERS.include? key
	end
      end
    end


    # This is the class for the most common type of look-up, an ItemSearch.
    #
    class ItemSearch < Operation

      # Not all search indices work in all locales. It is the user's
      # responsibility to ensure that a given index is valid within a given
      # locale.
      #
      # According to AWS documentation:
      #
      # *All* searches through all indices.
      # *Blended* combines DVD, Electronics, Toys, VideoGames, PCHardware,
      #   Tools, SportingGoods, Books, Software, Music, GourmetFood, Kitchen
      #   and Apparel.
      # *Merchants* combines all search indices for a merchant given with
      #   MerchantId.
      # *Music* combines the Classical, DigitalMusic, and MusicTracks indices.
      # *Video* combines the DVD and VHS search indices.
      #
      SEARCH_INDICES = %w[
	    All
	    Apparel              Hobbies              PetSupplies
            Automotive           HomeGarden           Photo
            Baby                 Jewelry              Software
            Beauty               Kitchen              SoftwareVideoGames
            Blended              Magazines            SportingGoods
            Books                Merchants            Tools
            Classical            Miscellaneous        Toys
            DigitalMusic         Music                VHS
            DVD                  MusicalInstruments   Video
            Electronics          MusicTracks          VideoGames
            ForeignBooks         OfficeProducts       Wireless
            GourmetFood          OutdoorLiving        WirelessAccessories
            HealthPersonalCare   PCHardware
	]


      def initialize(search_index, parameters)
	raise unless SEARCH_INDICES.include? search_index
	parameter_check( parameters )
	super( { 'SearchIndex' => search_index }.merge( parameters ) )
      end
    end


    # This is the class of look-up that deals with searching for items by
    # some attribute, such as the *ASIN* (Amazon Standard Item Number).
    #
    class ItemLookup < Operation
      def initialize(id_type, parameters, *b_parameters)

	id_type_str = 'IdType'

	unless b_parameters.empty?
	  class_str = self.class.to_s.sub( /^.+::/, '' )
	  id_type_str = '%s.Shared.IdType' % [ class_str ]
	  parameters = batch_parameters( parameters, *b_parameters )
	end

	super( { id_type_str => id_type }.merge( parameters ) )
      end
    end


    class SellerListingSearch < Operation
      def initialize(seller_id, parameters)
	super( { 'SellerId' => seller_id }.merge( parameters ) )
      end
    end


    class ListSearch < Operation
      def initialize(list_type, parameters)
	super( { 'ListType' => list_type}.merge( parameters ) )
      end
    end


    class BrowseNodeLookup < Operation
      def initialize(node, parameters)
	super( { 'BrowseNodeId' => node }.merge( parameters ) )
      end
    end



    # For batch request.
    #
    # class BatchOperation < Operation

    #   def initialize(*operations)
    #     raise if operations.any? { |op| op.kind != operations[0].kind }
    #   end

    # end


    # class MultiOperation

    #   def initialize(*operations)

    #     # The maximum number of operations in a multiple operation is 2.
    #     #
    #     raise if operations.size > 2

    #     op_names = operations.collect { |op| op.operation }.join( ',' )


    #   end
    # end
	

    class ResponseGroup

      RESPONSE_GROUPS = %w[
	BrowseNodeInfo BrowseNodes Large Medium Small Offers OfferFull
      ]

      # rg is just an alias for response_group.
      #
      attr_reader :response_group, :rg, :params

      def initialize(*rg)
	@response_group = @rg = rg
	@params = { 'ResponseGroup' => @rg.join( ',' ) }
      end
    end


    class OperationRequest < AWSObject

      attr_reader :arguments, :http_headers, :request_id,
		  :request_processing_time

      def initialize(xml)

	# Check for errors inside Request.
	#
	error_check( xml.elements['*/*/Request'] )

        root = xml.elements['*/OperationRequest']

	# Fundamental errors happen at the OperationRequest level. For
	# example, if an invalid AWSAccessKeyId is used.
	#
	error_check( root )

        @request_id = root.elements['RequestId'].text
        @request_processing_time =
	  root.elements['RequestProcessingTime'].text

        @http_headers = {}
        root.elements['HTTPHeaders'].elements.each do |e|
          @http_headers[e.attributes['Name']] = e.attributes['Value']
	end

        @arguments = {}
        root.elements['Arguments'].elements.each do |e|
          @arguments[e.attributes['Name']] = e.attributes['Value']
	end
      end

      def error_check(xml)
	if xml = xml.elements['Errors/Error']
	  e = AWSError.new( xml )
	  raise e.exception, e.msg
	end
      end

    end


    class ItemSet < AWSObject

      attr_reader :item_requests, :items, :search_bin_sets

      def initialize(xml)

	@item_requests = []
	@items = []

	# There may be more than one Items tag in the XML. This is the case,
	# for example, when a batch request is sent.
	#
	xml.get_elements( '*/Items' ).each do |item_set|
	  @item_requests << RequestStatus.new( item_set )

	  @items << item_set.get_elements( 'Item' ).collect do |e|
	    $stderr.print '.' if $DEBUG
	    Item.new( e )
	  end

	  if item_set.elements['SearchBinSets']
	    @search_bin_sets =
	      item_set.get_elements(
		'SearchBinSets/SearchBinSet' ).collect do |e|
	          SearchBinSet.new( e )
	    end
	  end

	end
	Amazon.dprintf

	@items.flatten!
      end

    end


    class SellerListingSet < AWSObject

      attr_reader :seller_listing_requests, :seller_listings

      def initialize(xml)
	@seller_listing_requests = []
	@seller_listings = []

        xml.get_elements( '*/SellerListings' ).each do |sl_set|
          @seller_listing_requests << RequestStatus.new( sl_set )

          @seller_listings = sl_set.get_elements( 'SellerListing' ).collect do |sl|
            $stderr.print '.' if $DEBUG
            SellerListing.new( sl )
          end

        end
        Amazon.dprintf

        @seller_listings.flatten!
      end
    end


    class SellerListing < AWSObject
      def initialize(xml)
	instantiate_attributes( xml )
      end
    end


    class Seller < AWSObject
      attr_reader :seller_id, :nickname

      def initialize(xml)

	@seller_id = xml.elements['SellerId'].text
	@nickname = xml.elements['Nickname'].text
      end

    end


    class BrowseNodeLookupSet < AWSObject

      attr_reader :browse_node_requests, :browse_nodes
      def initialize(xml)

	@browse_node_requests = []
	@browse_nodes = []

	xml.get_elements( '*/BrowseNodes' ).each do |bn_set|
	  @browse_node_requests << RequestStatus.new( bn_set )

	  @browse_nodes = bn_set.get_elements( 'BrowseNode' ).collect do |bn|
	    $stderr.print '.' if $DEBUG
	    BrowseNode.new( bn )
	  end

	end
	Amazon.dprintf

	@browse_nodes.flatten!
      end

    end


    class RequestStatus < AWSObject

      attr_reader :validity, :total_results, :total_pages

      def initialize(xml)
	root = xml

	@validity = root.elements['Request/IsValid'].text

	# Check to see if there were errors. If so, we take the error code
	# returned by AWS and create an exception class for it. If there
	# were multiple errors, we use only the first one, as we can only
	# raise a single exception.
	#
	if @validity == 'False'
	  e = AWSError.new( root.elements['Request/Errors/Error'] )
	  raise e.exception, e.msg
	end

	if new_root = root.elements['TotalResults']
	  @total_results = new_root.text
	end

	if new_root = root.elements['TotalPages']
	  @total_pages = new_root.text
	end
      end

    end


    class Item < AWSObject

      attr_reader :asin, :detail_page_url, :sales_rank,
		  :small_image, :medium_image, :large_image, :image_sets, 
		  :offer_summary, :offer_set, :customer_reviews,
		  :editorial_reviews, :similar_products, :accessories,
		  :browse_nodes, :listmania_lists

      def initialize(xml)
	root = xml

	@asin = root.elements['ASIN'].text

	if e = root.elements['DetailPageURL']
	  @detail_page_url = e.text
	end

	if e = root.elements['SalesRank']
	  @sales_rank = e.text
	end

	if new_root = root.elements['SmallImage']
	  @small_image = Image.new( new_root )
	end

	if new_root = root.elements['MediumImage']
	  @medium_image = Image.new( new_root )
	end

	if new_root = root.elements['LargeImage']
	  @large_image = Image.new( new_root )
	end

	if new_root = root.elements['ImageSets']

	  if new_root.elements['MerchantId']
	    @merchant_id = new_root.elements['MerchantId'].text
	  end

	  new_root = root.get_elements( 'ImageSets/ImageSet' )

	  @image_sets = []
	  new_root.each do |e|
	    @image_sets << ImageSet.new( e )
	  end
	end

	if new_root = root.elements['ItemAttributes']
	  instantiate_attributes( new_root )
	end

	if new_root = root.elements['OfferSummary']
	  @offer_summary = OfferSummary.new( new_root )
	end

	if new_root = root.elements['Offers']
	  @offer_set = OfferSet.new( new_root )
	end

	if new_root = root.elements['Tracks']
	  @tracks = TrackSet.new( new_root )
	end

	if new_root = root.elements['CustomerReviews']
	  @customer_reviews = CustomerReviewSet.new( new_root )
	end

	if new_root = root.elements['EditorialReviews']
	  @editorial_reviews = EditorialReviewSet.new( new_root )
	end

	if new_root = root.elements['SimilarProducts']
	  @similar_products = SimilarProductSet.new( new_root )
	end

	if new_root = root.elements['Accessories']
	  @accessories = AccessorySet.new( new_root )
	end

	if new_root = root.elements['BrowseNodes']
	  @browse_nodes = BrowseNodeSet.new( new_root )
	end

	if new_root = root.elements['ListmaniaLists']
	  @listmania_lists = ListmaniaListSet.new( new_root )
	end

      end

    end


    class Image < AWSObject

      attr_reader :url, :height, :height_units, :width, :width_units

      def initialize(xml)
	root = xml

	@url = root.elements['URL'].text
	@height = root.elements['Height'].text
	@height_units = root.elements['Height'].attributes['Units']
	@width = root.elements['Width'].text
	@width_units = root.elements['Width'].attributes['Units']
      end

    end


    class ImageSet < AWSObject

      attr_reader :category, :swatch_image, :small_image, :medium_image,
		  :large_image, :thumbnail_image, :tiny_image

      def initialize(xml)
	root = xml

	@category = root.attributes['Category']

	if new_root = root.elements['SwatchImage']
	  @swatch_image = Image.new( new_root )
	end

	if new_root = root.elements['SmallImage']
	  @small_image = Image.new( new_root )
	end
	
	if new_root = root.elements['MediumImage']
	  @medium_image = Image.new( new_root )
	end
	
	if new_root = root.elements['LargeImage']
	  @large_image = Image.new( new_root )
	end

	if new_root = root.elements['ThumbnailImage']
	  @thumbnail_image = Image.new( new_root )
	end

	if new_root = root.elements['TinyImage']
	  @tiny_image = Image.new( new_root )
	end
      end
    end


    class Price < AWSObject

      # attr_reader :amount, :currency_code, :formatted_price

      def initialize(xml)

	# Amount may be missing if price is 'Too low to display'.
	#
	if new_root = xml.elements['Amount']
	  @amount = new_root.text
	end

	# Currency code may be missing if price is 'Too low to display'.
	#
	if new_root = xml.elements['CurrencyCode']
	  @currency_code = new_root.text
	end

	@formatted_price = xml.elements['FormattedPrice'].text
      end
    end


    class Dimensions < AWSObject

      attr_reader :height, :height_units, :width, :width_units,
		  :length, :length_units, :weight, :weight_units

      def initialize(xml)
	root = xml

	if new_root = root.elements['Height']
	  @height = new_root.text
	  @height_units = new_root.attributes['Units']
	end

	if new_root = root.elements['Width']
	  @width = new_root.text
	  @width_units = new_root.attributes['Units']
	end

	if new_root = root.elements['Length']
	  @length = new_root.text
	  @length_units = new_root.attributes['Units']
	end

	if new_root = root.elements['Weight']
	  @weight = new_root.text
	  @weight_units = new_root.attributes['Units']
	end
      end
    end


    class ItemDimensions < Dimensions; end


    class PackageDimensions < Dimensions; end


    class OfferSummary < AWSObject

      attr_reader :items

      def initialize(xml)

	root = xml

	root.elements.each do |e|

	  if e.has_elements?
	    value = Price.new( e )
	  else
	    value = e.text
	  end

	  iv = Amazon.uncamelise( e.name )
	  instance_variable_set( "@#{iv}".to_sym, value )
	end

      end
    end


    class OfferSet < AWSObject

      attr_reader :total_offers, :total_offer_pages, :offers

      def initialize(xml)
	@total_offers = xml.elements['TotalOffers'].text
	@total_offer_pages = xml.elements['TotalOfferPages'].text
	@offers = xml.get_elements( 'Offer' ).collect { |e| Offer.new( e ) }
      end
    end


   class Offer < AWSObject

      attr_reader :merchant, :offer_attributes, :offer_listing

      def initialize(xml)
	@merchant = Merchant.new( xml.elements['Merchant'] )

	@offer_attributes =
	  OfferAttributes.new( xml.elements['OfferAttributes'] )

	@offer_listing = OfferListing.new( xml.elements['OfferListing'] )

      end

    end


    class TrackSet < Array

      def initialize(xml)
	super()
	self.concat( xml.get_elements( 'Disc' ).collect { |e| Disc.new( e ) } )
      end

    end


    class Disc < AWSObject

      attr_reader :number, :track

      def initialize(xml)
	@number = xml.attributes['Number']
	@track = xml.get_elements( 'Track' ).collect { |e| e.text }
      end

    end


    class Merchant < AWSObject

      attr_reader :merchant_id, :glance_page

      def initialize(xml)
	@merchant_id = xml.elements['MerchantId'].text
	@glance_page = xml.elements['GlancePage'].text
      end

    end


    class OfferAttributes < AWSObject

      attr_reader :condition, :sub_condition

      def initialize(xml)
	root = xml

	@condition = root.elements['Condition'].text

	if new_root = root.elements['SubCondition']
	  @sub_condition = new_root.text
	end
      end

    end


    class OfferListing < AWSObject

      attr_reader :offer_listing_id, :price, :availability,
		  :availability_attributes, :quantity, :super_saver

      def initialize(xml)
	root = xml

	@offer_listing_id = root.elements['OfferListingId'].text
	@price = Price.new( root.elements['Price'] )

	if new_root = root.elements['Availability']
	  @availability = root.elements['Availability'].text
	elsif new_root = root.elements['AvailabilityAttributes']
	  @availability_attributes = AvailabilityAttributes.new( new_root )
	end

	if new_root = root.elements['Quantity']
	  @quantity = new_root.text
	end

	@super_saver = root.elements['IsEligibleForSuperSaverShipping'].text
      end

    end


    class AvailabilityAttributes < AWSObject

      attr_reader :availability_type, :minimum_hours, :maximum_hours

      def initialize(xml)
	@availability_type = xml.elements['AvailabilityType'].text
	@minimum_hours = xml.elements['MinimumHours'].text
	@maximum_hours = xml.elements['MaximumHours'].text
      end

    end


    class ReviewSet < AWSObject

      attr_reader :reviews

      def initialize(xml, type='Review')
	@reviews = xml.get_elements( type ).collect { |e| Review.new( e ) }
      end
    end


    class CustomerReviewSet < ReviewSet

      attr_reader :average_rating, :total_reviews, :total_review_pages

      def initialize(xml)
	root = xml

	@average_rating = root.elements['AverageRating'].text
	@total_reviews = root.elements['TotalReviews'].text
	@total_review_pages = root.elements['TotalReviewPages'].text

	super( root )
      end
    end


    class EditorialReviewSet < ReviewSet

      def initialize(xml)
	super( xml, 'EditorialReview' )
      end
    end


    class Review < AWSObject

      def initialize(xml)
	root = xml

	root.elements.each do |e|
	  iv = Amazon.uncamelise( e.name )
	  
	  # Create a reader for the attribute.
	  #
	  Amazon::AWS.module_eval { attr_reader iv.to_sym }

	  instance_variable_set( "@#{iv}".to_sym, e.text )
	end

      end
    end


    class SimilarProductSet < Array

      def initialize(xml)
	super()
	self.concat( xml.collect { |e| SimilarProduct.new( e ) } )
      end

    end


    class SimilarProduct < AWSObject

      attr_reader :asin, :title

      def initialize(xml)
        @asin = xml.elements['ASIN'].text
        @title = xml.elements['Title'].text
      end

    end


    class AccessorySet < Array

      def initialize(xml)
	super()
	self.concat( xml.collect { |e| Accessory.new( e ) } )
      end

    end


    class Accessory < SimilarProduct

      def initialize(xml)
        super
      end

    end


    class LanguageSet < Array

      def initialize(xml)
	super()
	self.concat( xml.collect { |e| Language.new( e ) } )
      end

    end


    class Language < AWSObject

      attr_reader :name, :type, :audio_format

      def initialize(xml)
	@name = xml.elements['Name'].text
	@type = xml.elements['Type'].text

	if new_root = xml.elements['AudioFormat']
	  @audio_format = new_root.text
	end
      end

    end


   class BrowseNodeSet < Array

      def initialize(xml)
	super()
	self.concat( xml.collect { |e| BrowseNode.new( e ) } )
      end

    end


    class BrowseNode < AWSObject

      attr_reader :browse_node_id, :name, :ancestors, :children

      def initialize(xml)
	@browse_node_id = xml.elements['BrowseNodeId'].text

	if e = xml.elements['Name']
	  @name = e.text
	end

	if new_root = xml.elements['Ancestors']
	  @ancestors = BrowseNodeSet.new( new_root )
	end

	if new_root = xml.elements['Children']
	  @children = BrowseNodeSet.new( new_root )
	end
      end

    end


    class ListmaniaListSet < Array

      def initialize(xml)
	super()
	self.concat( xml.collect { |e| ListmaniaList.new( e ) } )
      end

    end


    class ListmaniaList < AWSObject

      attr_reader :list_id, :list_name

      def initialize(xml)
	@list_id = xml.elements['ListId'].text
	@list_name = xml.elements['ListName'].text
      end

    end


    # Note: Search bins are only available in the US locale.
    #

    class SearchBinSet < AWSObject

      attr_reader :narrow_by, :bins

      def initialize(xml)
	@narrow_by = xml.attributes['NarrowBy']
	@bins = xml.get_elements( 'Bin' ).collect { |e| Bin.new( e ) }
      end

    end


    class Bin < AWSObject

      attr_reader :bin_name, :bin_item_count, :bin_parameter

      def initialize(xml)
	@bin_name = xml.elements['BinName'].text
	@bin_item_count = xml.elements['BinItemCount'].text
	@bin_parameter = {
	  'Name'  => xml.elements['BinParameter/Name'].text,
	  'Value' => xml.elements['BinParameter/Value'].text
	}
      end

    end


    class ListSet < AWSObject

      attr_reader :list_search_request, :lists

      def initialize(xml)
	new_root = xml.elements['*/Lists']

	@list_search_request = RequestStatus.new( new_root )

      	@lists = new_root.get_elements( 'List' ).collect do |list|
	  $stderr.print '.' if $DEBUG
	  List.new( list )
	end

	Amazon.dprintf

      end

    end


    class List < AWSObject

      attr_reader :list_id

      def initialize(xml)
	instantiate_attributes(xml)
      end

    end

  end

end
